/*
 * iparapheur Core
 * Copyright (C) 2018-2025 Libriciel-SCOP
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: AGPL-3.0-only
 */

package coop.libriciel.ipcore.business.auth;

import coop.libriciel.ipcore.model.auth.Desk;
import coop.libriciel.ipcore.model.database.Subtype;
import coop.libriciel.ipcore.model.database.Tenant;
import coop.libriciel.ipcore.model.database.Type;
import coop.libriciel.ipcore.model.permission.DelegationDto;
import coop.libriciel.ipcore.services.auth.AuthServiceInterface;
import coop.libriciel.ipcore.services.database.SubtypeRepository;
import coop.libriciel.ipcore.services.database.TypeRepository;
import coop.libriciel.ipcore.utils.LocalizedStatusException;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.lang3.StringUtils;
import org.jetbrains.annotations.NotNull;
import org.springframework.stereotype.Service;

import java.util.*;

import static org.springframework.http.HttpStatus.*;


@Log4j2
@Service
public class DelegationBusinessService {


    // <editor-fold desc="Beans">


    private final AuthServiceInterface authService;
    private final SubtypeRepository subtypeRepository;
    private final TypeRepository typeRepository;


    public DelegationBusinessService(AuthServiceInterface authService,
                                     SubtypeRepository subtypeRepository,
                                     TypeRepository typeRepository) {
        this.authService = authService;
        this.subtypeRepository = subtypeRepository;
        this.typeRepository = typeRepository;
    }


    // </editor-fold desc="Beans">


    public void checkDelegationIntegrity(@NotNull Tenant tenant, @NotNull Desk desk, DelegationDto delegationDto) {

        if (StringUtils.equals(desk.getId(), delegationDto.getSubstituteDeskId())) {
            throw new LocalizedStatusException(NOT_ACCEPTABLE, "message.redundant_delegation_target");
        }

        log.debug("tenant:{}, desk:{}", tenant.getId(), delegationDto.getSubstituteDeskId());

        Optional.ofNullable(delegationDto.getSubstituteDeskId())
                .map(id -> authService.findDeskByIdNoException(tenant.getId(), id))
                .map(Desk::getId)
                .orElseThrow(() -> new LocalizedStatusException(NOT_FOUND, "message.unknown_desk_id"));

        if (StringUtils.isEmpty(delegationDto.getTypeId()) && StringUtils.isNotEmpty(delegationDto.getSubtypeId())) {
            log.error("The type is mandatory on a subtype-restricted delegation");
            throw new LocalizedStatusException(NOT_FOUND, "messages.unknown_subtype_id");
        }

        if (StringUtils.isNotEmpty(delegationDto.getTypeId()) && StringUtils.isEmpty(delegationDto.getSubtypeId())) {
            typeRepository.findByIdAndTenant_Id(delegationDto.getTypeId(), tenant.getId())
                    .map(Type::getId)
                    .orElseThrow(() -> new LocalizedStatusException(NOT_FOUND, "messages.unknown_type_id"));
        }

        if (StringUtils.isNoneEmpty(delegationDto.getTypeId(), delegationDto.getSubtypeId())) {
            subtypeRepository.findByIdAndParentType_IdAndTenant_Id(delegationDto.getSubtypeId(), delegationDto.getTypeId(), tenant.getId())
                    .map(Subtype::getId)
                    .orElseThrow(() -> new LocalizedStatusException(NOT_FOUND, "messages.unknown_subtype_id"));
        }
    }


}
